from base64 import b64encode, b64decode
from Crypto.Cipher import AES
from django.contrib.auth.hashers import make_password, check_password
from django.core.exceptions import ValidationError
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render, redirect

from dialog.settings import ENCRYPTION_KEY
from messages.auth import require_guest, require_login
from messages.models import Message, User


def main(request):
    return redirect('inbox')


@require_guest
def register_page(request):
    error = None

    if request.method == 'POST':
        try:
            username = request.POST.get('username', '')
            password = request.POST.get('password', '')
        
            if not username or not password:
                raise ValidationError('All fields are required.')

            try:
                user = User.objects.get(username=username)
                raise ValidationError('Username is already taken.')
            except User.DoesNotExist:
                pass

            if len(username) > 20:
                raise ValidationError('Username is too long.')
            if len(password) < 6:
                raise ValidationError('Password is too short.')

            password = make_password(password)

            user = User.objects.create(username=username, password=password)
            request.session['user_id'] = user.pk
            
            return redirect('inbox')
        except ValidationError as e:
            error = '\n'.join(e)

    return render(request, 'guest.html', {'register': True, 'error': error})


@require_guest
def login_page(request):
    error = None

    if request.method == 'POST':
        try:
            username = request.POST.get('username', '')
            password = request.POST.get('password', '')

            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist as e:
                raise ValidationError('Invalid username or password.') from e
            
            if check_password(password, user.password):
                request.session['user_id'] = user.pk
                return redirect('inbox')
            else:
                raise ValidationError('Invalid username or password.')
        except ValidationError as e:
            error = '\n'.join(e)
    return render(request, 'guest.html', {'register': False, 'error': error})


@require_login
def logout_page(request):
    response = HttpResponseRedirect('/login')
    response.delete_cookie('sessionid')
    return response


@require_login
def profile(request, user_id):
    user = get_object_or_404(User, pk=user_id)
    return render(request, 'profile.html', context={'page': 'profile', 'user': request.user, 'target_user': user})


@require_login
def compose_page(request):
    error = None
    if request.GET.get('to', '').isnumeric():
        target_id = request.GET['to']
    else:
        target_id = ''

    if request.method == 'POST':
        try:
            if request.user.is_blocked:
                raise ValidationError('This user can\'t send messages.')
            key = ENCRYPTION_KEY.encode()
            cipher = AES.new(key, AES.MODE_EAX)
            nonce = cipher.nonce

            user_id = request.POST.get('user_id', '')
            message = request.POST.get('message', '')

            if not user_id or not message:
                raise ValidationError('All fields are required.')

            if len(message) > 140:
                raise ValidationError('Message can\'t be longer than 140 characters.')

            try:
                user = User.objects.get(pk=int(user_id))
            except User.DoesNotExist as e:
                raise ValidationError('User does not exist.') from e

            ciphertext, tag = cipher.encrypt_and_digest(message.encode())
            message = '|'.join(map(bytes.decode, map(b64encode, [nonce, ciphertext, tag])))

            Message.objects.create(sender=request.user, recipient=user, text=message)

            return redirect('outbox')
        except ValidationError as e:
            error = '\n'.join(e)

    return render(request, 'compose.html', context={'page': 'compose', 'user': request.user, 'error': error, 'target_id': target_id})


@require_login
def inbox(request):
    messages = request.user.received.order_by('-created')[:10]
    return render(request, 'messages.html', context={'page': 'inbox', 'user': request.user, 'messages': messages})


@require_login
def outbox(request):
    messages = request.user.sent.order_by('-created')[:10]
    return render(request, 'messages.html', context={'page': 'outbox', 'user': request.user, 'messages': messages})

